import concurrent.futures
from multiprocessing import Manager
from pathlib import Path

from rich.progress import Progress

from plateaukit import parallel


class ParallelWriter:
    def __init__(self, writer_cls):
        self.writer = writer_cls()

    def transform(
        self,
        readable,
        outfile: str,
        *,
        split=2,
        progress_messages: dict | None = None,
        **kwargs,
    ):
        progress_messages = progress_messages or {}

        batches = readable.split(split)

        writer = self.writer

        with Progress() as rprogress:
            overall_progress_description = progress_messages.get(
                "description", "Processing..."
            )
            overall_task_id = rprogress.add_task(overall_progress_description)

            with Manager() as manager:
                quit = manager.Event()
                _progress = manager.dict()

                with concurrent.futures.ProcessPoolExecutor(max_workers=None) as pool:
                    futures = []
                    futures_status = dict()
                    for i, batch in enumerate(batches):
                        stem = Path(outfile).stem
                        if split > 1:
                            batch_outfile = Path(outfile).with_stem(f"{stem}.{i + 1}")
                        else:
                            batch_outfile = outfile

                        # task_id = rprogress.add_task(
                        #     f"[cyan]Progress #{i + 1}", total=len(batch)
                        # )

                        # logger.debug(f"batch_outfile: {batch_outfile}")

                        future = pool.submit(
                            writer.write_to,
                            batch,
                            batch_outfile,
                            **kwargs,
                        )

                        futures.append(future)
                        futures_status[future] = {
                            "task_id": i + 1,
                            "counter": i + 1,
                            "failed": False,
                        }

                    parallel.wait_futures(
                        futures,
                        pool,
                        quit,
                        futures_status,
                        overall_progress={
                            "task_id": overall_task_id,
                            "description": overall_progress_description,
                        },
                        rich_progress=rprogress,
                        shared_progress_status=_progress,
                    )



from __future__ import annotations

import os
import sys
from copy import deepcopy
from os import PathLike
from pathlib import Path
from typing import Optional

from plateaukit.config import Config, _get_data_items, _sort_dataset_entry
from plateaukit.download import downloader


def is_dataset_installed(dataset_id, format):
    config = Config()
    path = config.datasets.get(dataset_id, {}).get(format)
    return True if path else False
    # return path and Path(path).exists()


def list_installed_datasets():
    config = Config()
    return list(config.datasets.keys())


def install_dataset(
    dataset_id: str,
    format: str = "citygml",
    local: str | PathLike | None = None,
    force: bool = False,
):
    """Download and install PLATEAU datasets."""

    from plateaukit.download import city_list

    if not dataset_id:
        raise RuntimeError("Missing dataset_id")

    city = next(filter(lambda x: x["dataset_id"] == dataset_id, city_list), None)

    if not city:
        raise RuntimeError("Invalid dataset name")

    installed = is_dataset_installed(dataset_id, format)

    # Abort if a dataset is already installed
    if not force and installed:
        raise RuntimeError(
            f'ERROR: Dataset "{dataset_id}" ({format}) is already installed'
        )

    config = Config()

    current_entry = config.datasets[dataset_id] or dict()
    new_entry = deepcopy(current_entry)

    # COMPAT: If dataset_id entry is not empty and missing _spec, set _spec to 2
    if len(current_entry.keys()) > 0 and "_spec" not in current_entry:
        current_entry["_spec"] = "2"

    # If spec is mismatched, raise an error
    if "_spec" in current_entry and current_entry["_spec"] != city["spec"]:
        raise RuntimeError(
            f"Spec version mismatch: Installing v{city['spec']}, but v{current_entry['_spec']} is already installed."
            + f" Uninstall `{dataset_id}` first and try again."
        )
    else:
        new_entry["_spec"] = city["spec"]

    if local:
        local = Path(local).resolve()

        if not local.exists():
            raise RuntimeError("Local file not found")

        new_entry[format] = str(local)
        config.datasets[dataset_id] = _sort_dataset_entry(new_entry)
        config.save()

        return
    else:
        # TODO: Check beforehand if specified format is available in city_list
        resource_id = city.get(format, None)

        if not resource_id:
            raise RuntimeError(f"Format '{format}' is not available for '{dataset_id}'")

        config = Config()
        destfile_path = downloader.download_resource(resource_id, dest=config.data_dir)

        new_entry[format] = destfile_path
        config.datasets[dataset_id] = _sort_dataset_entry(new_entry)
        config.save()

        return


def get_dataset_filepaths(dataset_id: str, formats: list[str]):
    config = Config()
    paths = []
    for format in formats:
        if format == "parquet":
            record = config.datasets[dataset_id]["parquet"]
            if isinstance(record, dict):
                for path in config.datasets[dataset_id]["parquet"].values():
                    paths.append(path)
            elif isinstance(record, str):
                paths.append(record)
            else:
                raise RuntimeError("Invalid record for 'parquet'")
        else:
            path = config.datasets[dataset_id].get(format)
            if not path:
                raise RuntimeError(f"Missing files in record for '{format}'")
            paths.append(path)

    return paths


def uninstall_dataset(
    dataset_id: str, formats: Optional[list[str]] = None, keep_files: bool = False
):
    """Uninstall PLATEAU datasets."""

    config = Config()

    if dataset_id not in config.datasets:
        raise RuntimeError(f"Dataset '{dataset_id}' not found")

    formats = formats or list(_get_data_items(config.datasets[dataset_id]).keys())

    if not keep_files:
        paths = get_dataset_filepaths(dataset_id, formats)
        for path in paths:
            try:
                os.remove(path)
            except Exception as e:
                print(e, file=sys.stderr)

    for format in formats:
        if format in config.datasets[dataset_id]:
            del config.datasets[dataset_id][format]

    if len(_get_data_items(config.datasets[dataset_id]).items()) == 0:
        del config.datasets[dataset_id]

    config.save()



from bidict import bidict


class VerticesMap:
    counter: int
    index_by_vertex: bidict[tuple[float, float, float], int]

    def __init__(self):
        self.counter = 0
        self.index_by_vertex = bidict()

    def to_index(self, vertex: tuple[float, float, float]):
        if vertex not in self.index_by_vertex:
            self.index_by_vertex[vertex] = self.counter
            self.counter += 1
        return self.index_by_vertex[vertex]

    # def to_vertex(self, index):
    #     return self.index_by_vertex.inverse[index]

    # def exists_index(self, index):
    #     return index in self.index_by_vertex.inverse

    @property
    def vertices(self):
        return list(self.index_by_vertex.keys())

    def __contains__(self, vertex):
        return vertex in self.index_by_vertex



from __future__ import annotations

import click

from plateaukit.prebuild import prebuild
from plateaukit.logger import set_log_level


@click.command("prebuild")
@click.argument("dataset_id", nargs=1, required=True)
@click.option(
    "--format",
    type=click.Choice(
        ["parquet", "gpkg"],
        case_sensitive=False,
    ),
    default="parquet",
    help="Internal output format.",
)
@click.option(
    "--type",
    "-t",
    "types",
    type=click.Choice(
        ["bldg", "brid", "tran"],
        case_sensitive=True,
    ),
    default=["bldg"],
    multiple=True,
)
@click.option(
    "--split", default=10, help="Split the output into specified number of files."
)
@click.option("-v", "verbose", count=True, default=0, help="Verbose")
def prebuild_cmd(dataset_id: str, format, types: list[str], split: int, verbose: int):
    """Prebuild PLATEAU datasets.

    PLATEAU データセットを事前ビルドします。
    """

    if verbose >= 2:
        set_log_level("DEBUG")

    prebuild(
        dataset_id,
        format=format,
        types=types,
        split=split,
        simple_output=False if verbose else True,
    )



import warnings
from typing import Literal

from plateaukit.readers.citygml.ir_models import IRDocument
from plateaukit.readers.citygml.reader import Readable

from .geometry import GeometryTransformer


class LODFilteringTransformer(GeometryTransformer):
    def __init__(
        self,
        *,
        mode: Literal["highest", "all", "values"],
        values: list[str] | None = None,
    ):
        self.mode = mode

        if mode == "values" and values is None:
            raise ValueError("Values must be provided in 'values' mode.")

        self.values = values

    def transform(self, readable: Readable):
        readable.transformers.append(self)

        return readable

    def transform_document(self, document: IRDocument):
        if self.mode == "all":
            return document

        for city_object in document.city_objects:
            # TODO: Refactor this
            if not city_object.geometry:
                warnings.warn(f"CityObject {city_object.id} has no geometry.")
                continue

            if self.mode == "highest":
                max_lod_value = max(map(lambda x: x.lod, city_object.geometry))
                city_object.geometry = [
                    geom for geom in city_object.geometry if geom.lod == max_lod_value
                ]
            elif self.mode == "values":
                if self.values is None:
                    raise ValueError("Values must be provided in 'values' mode.")

                city_object.geometry = [
                    geom for geom in city_object.geometry if geom.lod in self.values
                ]
            else:
                raise ValueError(f"Invalid mode: {self.mode}")

        return document



import glob
import json
from pathlib import Path

import pandas as pd

from plateaukit.core.dataset import load_dataset


def _prebuild_cityobjects_db(
    dataset_id: str,
    outfile: str = "/tmp/cityobjects.parquet",
    types: list[str] = ["bldg", "tran", "brid"],
    tmpdir: str = "/tmp",
    split: int = 10,
):
    dataset = load_dataset(dataset_id)

    for type in types:
        outfile_cityjsonl = Path(tmpdir, f"{dataset_id}.{type}.city.jsonl")
        dataset.to_cityjson(
            outfile_cityjsonl,
            types=[type],
            seq=True,
            split=split,
            progress_messages={"description": f"Generating CityJSONSeq files: {type}"},
            # simple_output=simple_output,
        )

    # decoder = msgspec.json.Decoder(dict)

    buffer = []

    if split > 1:
        filename_pattern = str(Path(tmpdir, "*.city.*.jsonl"))
    else:
        filename_pattern = str(Path(tmpdir, "*.city.jsonl"))

    for type in types:
        for filename in glob.glob(filename_pattern):
            if f".{type}." not in filename:
                continue

            with open(filename, "r") as f:
                next(f)  # Skip the first line

                for i, line in enumerate(f):
                    cityjson = line.strip()
                    # data = decoder.decode(cityjson)
                    data = json.loads(cityjson)

                    objects = [
                        {
                            "_id": k,
                            "cityjson": cityjson,
                        }
                        for k in data["CityObjects"].keys()
                    ]
                    buffer += objects

    # TODO: Avoid having entire buffer in memory
    # print(buffer)
    # convert buffer to DataFrame. both `_id`, `cityjson` columns are str type
    df = pd.DataFrame(buffer, dtype=str)
    # print(df.head())
    df.to_parquet(outfile, compression="zstd")

    # return df



from plateaukit.readers.citygml.citygml_dataset import CityGMLDataset
from plateaukit.readers.citygml.parsers import (
    CityGMLParser,
    CityObjectParser,
    CodelistParser,
    PLATEAUCityGMLParser,
    PLATEAUCityObjectParser,
)

__all__ = [
    "CityGMLDataset",
    "CityGMLParser",
    "CityObjectParser",
    "CodelistParser",
    "PLATEAUCityGMLParser",
    "PLATEAUCityObjectParser",
]



from .filter_lod import LODFilteringTransformer
from .reprojection import ReprojectionTransformer
from .shift_to_ground import ShiftToGroundTransformer

__all__ = [
    "ShiftToGroundTransformer",
    "LODFilteringTransformer",
    "ReprojectionTransformer",
]



from typing import Any

from plateaukit.exporters.cityjson.vertices_map import VerticesMap
from plateaukit.readers.citygml.ir_models import IRDocument, IRGeometry, IRMetadata
from plateaukit.utils import dict_key_to_camel_case

# @dataclass
# class IndexedGeometry:
#     type: str
#     boundaries: list[Any]
#     lod: str | None = None


class CityJSONWriter:
    def _get_indexed_boundaries(self, geometry: IRGeometry, vertices_map: VerticesMap):
        """Get indexed boundaries from geometry."""

        # TODO: handling composite surface seriously
        # print("get_indexed_boundaries")
        # print("type", geometry.type)

        boundaries = geometry.boundaries
        indexed_boundaries: list[Any] = []  # TODO: typing

        if geometry.type == "MultiPoint":
            for point in geometry.boundaries:
                index = vertices_map.to_index(point)
                indexed_boundaries.append(index)
            return indexed_boundaries, vertices_map

        elif geometry.type == "MultiLineString":
            for line in geometry.boundaries:
                indexed_line = []
                for point in line:
                    index = vertices_map.to_index(point)
                    indexed_line.append(index)
                indexed_boundaries.append(indexed_line)
            return indexed_boundaries, vertices_map

        elif geometry.type in ["MultiSurface", "CompositeSurface"]:
            for surface in boundaries:
                # print("surface", surface)
                indexed_surface = []
                for region in surface:
                    # print("region", region)

                    # NOTE: Reverse the order of vertices; this is necessary for
                    # being rendered correctly in cityjson-threejs-loader
                    # for SOME datasets
                    # eg. plateau-13213-higashimurayama-shi-2020
                    ## region = region[::-1]

                    indexed_region = []
                    unclosed_region = region[:-1]
                    for point in unclosed_region:
                        index = vertices_map.to_index(point)
                        indexed_region.append(index)
                    indexed_surface.append(indexed_region)
                indexed_boundaries.append(indexed_surface)
            # print("indexed_boundaries", indexed_boundaries, "aaa" * 100)
            return indexed_boundaries, vertices_map

        elif geometry.type == "Solid":
            for shell in boundaries:
                # print("shell", shell)
                indexed_shell = []
                for surface in shell:
                    indexed_surface = []
                    for region in surface:
                        # print("region", region)
                        indexed_region: list[int] = []
                        unclosed_region = region[:-1]
                        for point in unclosed_region:
                            # print("point", point)
                            index = vertices_map.to_index(point)
                            indexed_region.append(index)
                        # print("indexed_region", indexed_region)
                        indexed_surface.append(indexed_region)
                    indexed_shell.append(indexed_surface)
                indexed_boundaries.append(indexed_shell)
            # print("indexed_boundaries", indexed_boundaries)
            return indexed_boundaries, vertices_map
        else:
            raise NotImplementedError()

        # TODO: MultiSolid, CompositeSolid

        # for surface in exterior:
        #     unclosed_surface = surface[:-1]
        #     # print("unclosed_surface", unclosed_surface)
        #     single_exterior_surface_exterior = []

        #     for chunk in unclosed_surface:
        #         index = vertices_map.to_index(chunk)
        #         single_exterior_surface_exterior.append(index)

        #     indexed_surface = [single_exterior_surface_exterior]
        #     indexed_surfaces.append(indexed_surface)

        # return indexed_surfaces, vertices_map

    def get_indexed_geometry(self, geometry: IRGeometry, vertices_map: VerticesMap):
        """Convert geometries to an indexed geometry with a vertices map."""

        indexed_boundaries, vertices_map = self._get_indexed_boundaries(
            geometry,
            vertices_map,
        )

        indexed_geometry = IRGeometry(
            type=geometry.type,
            lod=geometry.lod,
            boundaries=indexed_boundaries,
            semantics=geometry.semantics,
        )

        return indexed_geometry, vertices_map

    def _get_meta(self, metadata: IRMetadata) -> dict:
        return {
            "type": "CityJSON",
            "version": "2.0",
            # "extensions": {},
            "transform": {"scale": [1.0, 1.0, 1.0], "translate": [0.0, 0.0, 0.0]},
            "metadata": {
                "referenceSystem": f"https://www.opengis.net/def/crs/EPSG/0/{metadata.epsg}",
            },
            "vertices": [],
            # "appearance": {},
            # "geometry-templates": {},
        }

    def transform(
        self,
        document: IRDocument,
        *,
        seq=False,
    ) -> list | dict:
        if seq:
            lines = []
            cj_meta = self._get_meta(document.metadata)
            lines.append(cj_meta)

            for obj in document.city_objects:
                indexed_geometries = []
                vertices_map = VerticesMap()

                if obj.geometry is None:
                    pass
                else:
                    for geometry in obj.geometry:
                        indexed_geometry, vertices_map = self.get_indexed_geometry(
                            geometry, vertices_map
                        )
                        indexed_geometries.append(indexed_geometry)

                if obj.attributes is not None:
                    attributes = dict_key_to_camel_case(obj.attributes)
                else:
                    attributes = {}

                result = {
                    "type": "CityJSONFeature",
                    "CityObjects": {
                        obj.id: {
                            "type": obj.type,
                            "attributes": attributes,
                            "geometry": [
                                {
                                    "type": indexed_geometry.type,
                                    "lod": indexed_geometry.lod,
                                    "boundaries": indexed_geometry.boundaries,
                                }
                                for indexed_geometry in indexed_geometries
                            ],
                        }
                    },
                    "vertices": vertices_map.vertices,
                    # "appearance": {},
                    # "geometry-templates": {},
                }
                lines.append(result)

            return lines

        else:
            # raise NotImplementedError()
            vertices_map = VerticesMap()
            city_objects = {}

            for obj in document.city_objects:
                indexed_geometries = []

                for geometry in obj.geometry:
                    indexed_geometry, vertices_map = self.get_indexed_geometry(
                        geometry, vertices_map
                    )
                    indexed_geometries.append(indexed_geometry)

                city_objects.update(
                    {
                        obj.id: {
                            "type": obj.type,
                            "attributes": obj.attributes,
                            "geometry": [
                                {
                                    "type": indexed_geometry.type,
                                    "lod": indexed_geometry.lod,
                                    "boundaries": indexed_geometry.boundaries,
                                }
                                for indexed_geometry in indexed_geometries
                            ],
                        }
                    }
                )
            result = {
                "type": "CityJSON",
                "version": "2.0",
                "extensions": {},
                "transform": {"scale": [1.0, 1.0, 1.0], "translate": [0.0, 0.0, 0.0]},
                "metadata": {
                    "referenceSystem": f"https://www.opengis.net/def/crs/EPSG/0/{document.metadata.epsg}",
                },
                "CityObjects": city_objects,
                "vertices": vertices_map.vertices,
                # "appearance": {},
                # "geometry-templates": {},
            }

            return result

    def write_to(
        self,
        document_or_readable,
        outfile: str,
        *,
        seq=False,
    ):
        import json

        if hasattr(document_or_readable, "read"):
            document = document_or_readable.read()
        else:
            document = document_or_readable

        result = self.transform(document, seq=seq)

        with open(outfile, "w") as f:
            if seq:
                for line in result:
                    f.write(
                        json.dumps(line, separators=(",", ":"), ensure_ascii=False)
                        + "\n"
                    )
            else:
                f.write(json.dumps(result, separators=(",", ":"), ensure_ascii=False))



from __future__ import annotations

import click

from plateaukit.logger import set_log_level


# https://alexdelorenzo.dev/notes/click.html
class OrderCommands(click.Group):
    def list_commands(self, ctx):
        return list(self.commands)


# https://stackoverflow.com/a/76624494/10954858
class AliasCliGroup(click.Group):
    def command(self, *args, **kwargs):
        """Adds the ability to add `aliases` to commands."""

        def decorator(f):
            aliases = kwargs.pop("aliases", None)
            if aliases and isinstance(aliases, list):
                name = kwargs.pop("name", None)
                if not name:
                    raise click.UsageError(
                        "`name` command argument is required when using aliases."
                    )

                base_command = super(AliasCliGroup, self).command(
                    name, *args, **kwargs
                )(f)

                for alias in aliases:
                    cmd = super(AliasCliGroup, self).command(alias, *args, **kwargs)(f)
                    cmd.help = f"Alias for '{name}'.\n\n{cmd.help}"
                    cmd.params = base_command.params

            else:
                cmd = super(AliasCliGroup, self).command(*args, **kwargs)(f)

            return cmd

        return decorator


class CustomCliGroup(AliasCliGroup, OrderCommands):
    pass


@click.group(
    cls=CustomCliGroup,
    context_settings=dict(help_option_names=["-h", "--help"], show_default=True),
    no_args_is_help=True,
)
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="Enable verbose mode."
)
def cli(verbose):
    if verbose:
        set_log_level("DEBUG")



from dataclasses import dataclass

import pyproj
from lxml import etree

from plateaukit.readers.citygml.constants import default_nsmap
from plateaukit.readers.citygml.extractors import city_object_extractors as extractors
from plateaukit.readers.citygml.ir_models import IRCityObject, IRGeometry
from plateaukit.readers.citygml.parsers.geometry_parser import GeometryParser
from plateaukit.readers.citygml.parsers.xml_models.city_object import CityObjectXML


@dataclass
class Building(IRCityObject):
    address: dict | None = None


class CityObjectParser:
    """A parser for CityGML objects.

    Attributes:
        transformer: A pyproj.Transformer instance.
    """

    transformer: pyproj.Transformer | None
    nsmap: dict[str, str]

    def __init__(
        self,
        *,
        transformer: pyproj.Transformer | None = None,
        nsmap: dict[str, str],
        codelist_map: dict | None = {},
    ):
        self.transformer = transformer
        self.nsmap = nsmap
        self.codelist_map = codelist_map or {}

        self.object_type_tags = {
            "Building": f"{{{default_nsmap['bldg']}}}Building",
            "Road": f"{{{default_nsmap['tran']}}}Road",
            "Bridge": f"{{{default_nsmap['brid']}}}Bridge",
        }

        self.geometry_type_tags = [
            {"type": "lod0RoofEdge", "tag": f"{{{default_nsmap['bldg']}}}lod0RoofEdge"},
            {
                "type": "lod0FootPrint",
                "tag": f"{{{default_nsmap['bldg']}}}lod0FootPrint",
            },
            {"type": "lod1Solid", "tag": f"{{{default_nsmap['bldg']}}}lod1Solid"},
            {"type": "lod2Solid", "tag": f"{{{default_nsmap['bldg']}}}lod2Solid"},
            {
                "type": "lod2MultiSurface",
                "tag": f"{{{default_nsmap['bldg']}}}lod2MultiSurface",
            },
            {
                "type": "lod1MultiSurface",
                "tag": f"{{{default_nsmap['tran']}}}lod1MultiSurface",
            },
            {"type": "lod1Solid", "tag": f"{{{default_nsmap['brid']}}}lod1Solid"},
            {
                "type": "lod2MultiSurface",
                "tag": f"{{{default_nsmap['brid']}}}lod2MultiSurface",
            },
        ]


class PLATEAUCityObjectParser(CityObjectParser):
    """A parser for PLATEAU CityGML objects."""

    def _get_geometry(self, root: etree._Element):
        geoms = []

        parser = GeometryParser(transformer=self.transformer)

        for type_tag in self.geometry_type_tags:
            type = type_tag["type"]
            tag = type_tag["tag"]

            el = root.find(f"./{tag}", self.nsmap)

            if el is None:
                continue

            if type in ["lod0RoofEdge", "lod0FootPrint"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = IRGeometry(
                    type="MultiSurface",
                    lod="0",
                    boundaries=chunked_poslists,
                    semantics={
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                )

                geoms.append(geom)
            elif type in ["lod1MultiSurface"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = IRGeometry(
                    type="MultiSurface",
                    lod="1",
                    boundaries=chunked_poslists,
                    semantics={
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                )

                geoms.append(geom)
            elif type in ["lod1Solid"]:
                # TODO: Fix this
                chunked_poslists = parser.extract_chunked_poslists(el)
                solid_boundaries = [chunked_poslists]

                geom = IRGeometry(
                    type="Solid",
                    lod="1",
                    boundaries=solid_boundaries,
                    semantics={
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(solid_boundaries))],
                    },
                )

                geoms.append(geom)
            elif type in ["lod2MultiSurface"]:
                chunked_poslists = parser.extract_chunked_poslists(el)

                geom = IRGeometry(
                    type="MultiSurface",
                    lod="2",
                    boundaries=chunked_poslists,
                    semantics={
                        "surfaces": [
                            {
                                "type": f"+{type}",
                            }
                        ],
                        "values": [0 for _ in range(len(chunked_poslists))],
                    },
                )

                geoms.append(geom)

        # NOTE: Process boundedBy for Building and Bridge

        if root.tag in [
            self.object_type_tags["Building"],
            self.object_type_tags["Bridge"],
        ]:
            # Look through boundedBy
            # TODO: Fix these conditions
            if root.tag == self.object_type_tags["Building"] and "bldg" in self.nsmap:
                bound_els = list(root.iterfind("./bldg:boundedBy", self.nsmap))
            elif root.tag == self.object_type_tags["Bridge"] and "brid" in self.nsmap:
                bound_els = list(root.iterfind("./brid:boundedBy", self.nsmap))
            else:
                # TODO: Fix this
                raise NotImplementedError()
                # print("self.nsmap", self.nsmap)
                # exit()
                # bound_els = []

            for bound_el in bound_els:
                # TODO: Check semantics
                for type_tag in self.geometry_type_tags:
                    type = type_tag["type"]
                    tag = type_tag["tag"]

                    el = bound_el.find(f".//{tag}", self.nsmap)

                    if el is None:
                        continue

                    if type in ["lod2MultiSurface"]:
                        chunked_poslists = parser.extract_chunked_poslists(el)

                        geom = IRGeometry(
                            type="MultiSurface",
                            lod="2",
                            boundaries=chunked_poslists,
                            semantics={
                                "surfaces": [
                                    {
                                        "type": f"+{type}",
                                    }
                                ],
                                "values": [0 for _ in range(len(chunked_poslists))],
                            },
                        )

                        geoms.append(geom)

                    else:
                        pass
                        # raise NotImplementedError()
        # elif root.tag == self.object_type_tags["Road"]:
        #     pass
        # else:
        #     raise NotImplementedError(f"Failed to parse geometry for {root.tag}")

        return geoms

    def _parse_attributes(self, el: CityObjectXML) -> dict:
        # TODO: parse `uro:` attributes

        attributes = dict()

        if el.tree.tag == self.object_type_tags["Building"]:
            attributes["building_id"] = extractors.get_building_id(el)

            # Optional attributes
            # TODO: Change extractors to transformers
            optional_attributes = {
                "measured_height": extractors.get_measured_height(el),
                "year_of_construction": extractors.get_year_of_construction(el),
                "storeys_above_ground": extractors.get_storeys_above_ground(el),
                "storeys_below_ground": extractors.get_storeys_below_ground(el),
                "name": extractors.get_name(el),
                "usage": extractors.get_usage(el),
                "district_plan": extractors.get_district_plan(el),
            }

            # river_flooding_risks = extractors.get_river_flooding_risks(el)
            # river_flooding_risk_attributes = {}
            # for name, dic in river_flooding_risks.items():
            #     for item_key, value in dic.items():
            #         key = f"river_flooding_risk.{name}.{item_key}"
            #         river_flooding_risk_attributes[key] = value
            # optional_attributes.update(river_flooding_risk_attributes)

            # TODO: Delete column if all values are empty (elsewhere)

            river_flooding_risks = extractors.get_river_flooding_risks(el)
            optional_attributes.update({"river_flooding_risk": river_flooding_risks})

            districts_and_zones_type = extractors.get_districts_and_zones_type(el)
            optional_attributes.update(
                {"districts_and_zones_type": districts_and_zones_type}
            )

            optional_attributes = {
                k: v for k, v in optional_attributes.items() if v is not None
            }
            attributes.update(optional_attributes)
        elif el.tree.tag == self.object_type_tags["Bridge"]:
            attributes["_gml_id"] = el.get_gml_id()

        return attributes

    def parse(self, element: etree._Element) -> IRCityObject:
        el = CityObjectXML(element, nsmap=self.nsmap, codelist_map=self.codelist_map)

        citygml_id = el.get_gml_id()
        address = None  # self._get_address(el)

        attributes = self._parse_attributes(el)

        if el.tree.tag == self.object_type_tags["Building"]:
            attributes["building_id"] = extractors.get_building_id(el)

            geometry = self._get_geometry(el.tree)

            obj = Building(
                type="Building",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
                address=address,
            )
        elif el.tree.tag == self.object_type_tags["Road"]:
            geometry = self._get_geometry(el.tree)

            obj = IRCityObject(
                type="Road",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
            )
        elif el.tree.tag == self.object_type_tags["Bridge"]:
            geometry = self._get_geometry(el.tree)

            obj = IRCityObject(
                type="Bridge",
                id=citygml_id,
                attributes=attributes,
                geometry=geometry,
            )
        else:
            raise NotImplementedError(f"Unknown object type: {el.tree.tag}")
            # warnings.warn(f"Unknown object type: {el.tree.tag}")

        return obj



from pathlib import Path
from typing import TextIO

import pandas as pd

from plateaukit.config import Config
from plateaukit.logger import logger


def to_geojson(self, file: str | None = None):
    """Export the area in GeoJSON format."""

    # TODO: Support GeoJSONSeq

    data = self.gdf.to_json(ensure_ascii=False)

    if file is not None:
        with open(file, "w") as f:
            f.write(data)
    else:
        return data


def to_cityjson(
    self,
    file: str | TextIO,
    *,
    types: list[str] | None = None,
    ground: bool = False,
    seq: bool = False,
    target_epsg: int = 4326,
):
    """Export the area in CityJSON format."""

    # TODO: Support IOBase as file

    from plateaukit import Dataset

    if self._datasets is None:
        raise RuntimeError("Missing dataset information")

    types = list(self.layers.keys()) if types is None else types

    # TODO: Support non-building types
    # selection = self.gdf["buildingId"].tolist()
    selection = sum([layer.gdf["gmlId"].tolist() for layer in self.layers.values()], [])
    logger.debug(selection)

    config = Config()

    for dataset_id in self._datasets:
        # co_parquet_path = Path(config.data_dir, f"{dataset_id}.cityobjects.parquet")
        co_parquet_path = config.datasets.get(dataset_id, {}).get("cityobjects", None)
        co_parquet_path = Path(co_parquet_path) if co_parquet_path else None

        if seq and co_parquet_path is not None and target_epsg == 4326:
            selection = sum(
                [layer.gdf["gmlId"].tolist() for layer in self.layers.values()], []
            )
            df = pd.read_parquet(co_parquet_path)
            # Get rows where df._id in selection:
            df = df[df._id.isin(selection)]
            # Write value of `cityjson` row of each row to a line in the file.
            assert target_epsg == 4326
            cjseq_header = (
                '{"type":"CityJSON","version":"2.0","transform":{"scale":[1.0,1.0,1.0],"translate":[0.0,0.0,0.0]},'
                + f'"metadata":{{"referenceSystem":"https://www.opengis.net/def/crs/EPSG/0/{target_epsg}"}},"vertices":[]}}'
            )
            if isinstance(file, str):
                with open(file, "w") as f:
                    f.write(cjseq_header + "\n")
                    for row in df.itertuples():
                        f.write(str(row.cityjson) + "\n")
            else:
                file.write(cjseq_header + "\n")
                for row in df.itertuples():
                    file.write(str(row.cityjson) + "\n")
                file.seek(0)
        else:
            if not isinstance(file, str):
                raise ValueError("File-like object is not supported for direct export")

            dataset = Dataset(dataset_id)

            dataset.to_cityjson(
                file,
                types=types,
                ground=ground,
                selection=selection,
                target_epsg=target_epsg,
                seq=seq,
            )



from typing import Literal
import cityview as cv
import geopandas as gpd
import ipywidgets as widgets
from IPython.display import display

import plateaukit.core as core


class CityViewWidget:
    def __init__(
        self,
        area,
        *,
        mode: Literal["virtual", "map"] = "virtual",
        theme: Literal["light", "dark"] = "light",
    ) -> None:
        from io import StringIO

        sink = StringIO()
        area.to_cityjson(sink, target_epsg=4326, seq=True)
        cjseq_data = sink.read()

        self.cjseq_data = cjseq_data

        if mode == "map":
            self.cityview = cv.MapView(theme=theme)
        else:
            self.cityview = cv.VirtualView(theme=theme)

        layers = [cv.CityJSONLayer(data=cjseq_data, format="cityjsonseq")]
        self.cityview.layers = layers
        self.cityview.update()

        # Set selection area
        empty_gdf = gpd.GeoDataFrame()

        # TODO: Fix this; related to circular import
        layer = core.area.GeoDataFrameLayer(empty_gdf)

        # TODO: Fix this; related to circular import
        self.selection = core.area.Area(layer, base_layer_name="bldg")

        self._widget_df = area.gdf.copy()

        self.out = widgets.Output()

        def display_df():
            with self.out:
                display(self._widget_df, clear=True)

            return self.out

        # self.cityview.observe(self.click_handler, names="click")

        self.widget = widgets.VBox([self.cityview, self.out])

    def click_handler(self, change):
        pass
        # feature = change["new"]
        # print("clicked", feature)

        # if "fill_color" in feature["properties"]:
        #     del feature["properties"]["fill_color"]

        # pyogrio_invalid_chars_regex = re.compile(r"[;]")

        # for key, value in feature["properties"].items():
        #     if isinstance(value, str) and pyogrio_invalid_chars_regex.search(value):
        #         feature["properties"][key] = feature["properties"][key].replace(
        #             ";", ","
        #         )

        # # NOTE: https://github.com/heremaps/xyz-spaces-python/pull/115
        # buf = json.dumps(feature)
        # row = gpd.read_file(StringIO(buf))

        # if self.selection.gdf.empty:
        #     self.selection.gdf = row
        # else:
        #     # If self.selection.gdf contains the row, remove it
        #     if row.at[0, "buildingId"] in self.selection.gdf["buildingId"].values:
        #         self.selection.gdf = self.selection.gdf[
        #             ~self.selection.gdf.loc[:, "buildingId"].isin(row["buildingId"])
        #         ]
        #     else:
        #         self.selection.gdf = pd.concat([self.selection.gdf, row])

        # overlap_df = self._widget_df.loc[
        #     self._widget_df.loc[:, "buildingId"].isin(self.selection.gdf["buildingId"])
        # ]

        # # Add fill_color column to df if it doesn't exist
        # if "fill_color" not in self._widget_df.columns:
        #     self._widget_df.loc[:, "fill_color"] = None

        # self._widget_df["fill_color"] = self._widget_df.loc[
        #     self._widget_df.index.isin(overlap_df.index), "fill_color"
        # ].apply(lambda x: [255, 0, 0])

        # updated_layer = self._create_building_layer(self._widget_df)
        # self.deck.layers = [updated_layer]

        # self.deck.update()

        # # Delete fill_color column
        # self._widget_df.drop(columns=["fill_color"], inplace=True)

        # with self.out:
        #     # print("clicked", feature)
        #     display(
        #         self._widget_df.iloc[self._widget_df.index.isin(overlap_df.index)],
        #         clear=True,
        #     )
